from django.db import models
from django.contrib.auth.models import User
from django.core.validators import RegexValidator


telefon_regex = RegexValidator(
    regex=r"^\+\d{9,15}$",
    message="Telefonnummer muss von folgendem Format sein: '+4938149444614'. "
    "Es sind bis zu 15 Ziffern erlaubt."
)


class Studienabschnitt(models.Model):
    """
    z.B. "1. Semester", "2. Semester", "Physikum", "STEX", "PJ", ...
    """
    bezeichnung = models.CharField(max_length=100)
    sortierung = models.PositiveSmallIntegerField()

    def __str__(self):
        return self.bezeichnung

    class Meta:
        verbose_name = "Studienabschnitt/Semester"
        verbose_name_plural = "Studienabschnitte/Semester"
        ordering = ["sortierung"]


class Studiengang(models.Model):
    """
    z.B. "Humanmedizin", "Zahnmedizin", ...
    """
    bezeichnung = models.CharField(max_length=100, unique=True)
    studienabschnitt = models.ManyToManyField(
        Studienabschnitt,
        blank=True,
        related_name="studiengang"
    )

    def __str__(self):
        return self.bezeichnung

    class Meta:
        verbose_name = "Studiengang"
        verbose_name_plural = "Studiengänge"
        ordering = ["bezeichnung"]


class Gender(models.Model):
    bezeichnung = models.CharField(max_length=100)
    # z.B. "in" für "weiblich" -> "Studentin"
    endung = models.CharField(
        max_length=8,
        blank=True,
    )

    def __str__(self):
        return self.bezeichnung

    class Meta:
        verbose_name = "Gender/Geschlecht"
        verbose_name_plural = "Gender/Geschlechter"
        ordering = ["bezeichnung"]


class FachschaftUser(models.Model):
    """
    Extends django.contrib.auth.models.User
    FachschaftUser = "Student:innen allgemein"
    """
    user = models.OneToOneField(
        User,
        on_delete=models.PROTECT,
        related_name="fachschaftuser",
    )
    gender = models.ForeignKey(Gender, on_delete=models.PROTECT)
    studienabschnitt = models.ForeignKey(
        Studienabschnitt,
        on_delete=models.PROTECT,
    )
    studiengang = models.ForeignKey(
        Studiengang,
        on_delete=models.PROTECT,
    )
    nickname = models.CharField(
        max_length=100,
        unique=True,
    )

    def __str__(self):
        """
        returns "[Vorname] [Nachname]"
        """
        return self.user.get_full_name()

    class Meta:
        verbose_name = "Student:in"
        verbose_name_plural = "Student:innen"
        ordering = ["nickname"]


class Kontaktdaten(models.Model):
    """
    Die Kontaktdaten werden vom fachschaftUser ausgelagert, weil auch historische
    Kontaktdaten erfasst werden müssen. Wenn jemand die Kontaktdaten verifizieren
    lässt, dann ein Buch ausleiht und dann die Kontaktdaten ändert, dann weiß niemand
    mehr, wie das Buch wieder eingetrieben werden kann.
    """
    fachschaftuser = models.ForeignKey(
        FachschaftUser,
        on_delete=models.CASCADE,
        related_name="kontaktdaten",
    )
    created = models.DateTimeField(
        auto_now_add=True,
        verbose_name="erzeugt",
    )
    """
    Die Adresse ist einfach nur ein Textfeld, weil es von Menschen manuell überprüft
    werden soll. Daher spalten wir das nicht extra auf.
    """
    adresse = models.TextField()
    telefonnummer = models.CharField(
        max_length=16,
        default="",
        validators=[telefon_regex],
        verbose_name="Telefonnummer",
        help_text=(
            "Telefonnummer muss von folgendem Format sein: '+4938149444614'. "
            "Es sind bis zu 15 Ziffern erlaubt."
        ),
    )
    gecheckt_von = models.ForeignKey(
        User,
        on_delete=models.SET_NULL,
        blank=True,
        null=True,
        verbose_name="überprüft von",
    )
    gecheckt_datum = models.DateTimeField(
        blank=True,
        null=True,
        verbose_name="überprüft am",
    )

    class Meta:
        verbose_name = "Kontaktdaten"
        verbose_name_plural = "Kontaktdaten"
        ordering = ["-created"]

    def __str__(self):
        return f"{self.fachschaftuser} {self.created}"

    def ist_verifiziert(self):
        return all([self.gecheckt_von, self.gecheckt_datum])



from django.contrib.auth.models import User
from django.urls import path

from rest_framework import serializers, generics
from rest_framework.urlpatterns import format_suffix_patterns
from rest_framework.views import APIView
from rest_framework.response import Response
from rest_framework.permissions import DjangoModelPermissions
from rest_framework.exceptions import PermissionDenied

from .models import Kontaktdaten, FachschaftUser
from .apps import app_name as parent_app_name

app_name = 'api'

# Serializers

class UserSerializer(serializers.ModelSerializer):
    class Meta:
        model = User
        fields = [
            'id',
            'username',
            'first_name',
            'last_name',
        ]


class FachschaftUserSerializer(serializers.HyperlinkedModelSerializer):
    user = serializers.HyperlinkedRelatedField(
        read_only=True,
        view_name=f'{parent_app_name}:{app_name}:user-detail',
        lookup_field='id',
    )
    class Meta:
        model = FachschaftUser
        fields = [
            'id',
            'nickname',
            'user',
        ]


class KontaktdatenSerializer(serializers.HyperlinkedModelSerializer):
    fachschaftuser = serializers.HyperlinkedRelatedField(
        read_only=True,
        view_name=f'{parent_app_name}:{app_name}:fachschaftuser-detail',
        lookup_field='id',
    )
    gecheckt_von = serializers.HyperlinkedRelatedField(
        read_only=True,
        view_name=f'{parent_app_name}:{app_name}:user-detail',
        lookup_field='id',
    )
    class Meta:
        model = Kontaktdaten
        fields = [
            'fachschaftuser',
            'created',
            'adresse',
            'telefonnummer',
            'gecheckt_von',
            'gecheckt_datum',
        ]


# Views

class UserDetail(generics.RetrieveAPIView):
    queryset = User.objects.all()
    serializer_class = UserSerializer
    lookup_field = 'id'


class FachschaftUserDetail(generics.RetrieveAPIView):
    queryset = FachschaftUser.objects.all()
    serializer_class = FachschaftUserSerializer
    lookup_field = 'id'


class KontaktdatenListByFachschaftUser(APIView):
    permission_classes = [DjangoModelPermissions]
    queryset = Kontaktdaten.objects.all()

    def get(self, request, id, format=None):
        """
        Die normalen DjangoModelPermissions funktionieren nicht gut genug (jeder darf
        lesen), daher können wir die DjangoModelPermissions-Klasse ableiten oder
        einfach so checken, ob man die richtigen Permissions hat:

        (ohne permission_classes geht's aber auch nicht, daher ist das hier der
        einfachste Weg)
        """
        if not (request.user.has_perm(f'{parent_app_name}.view_kontaktdaten') or
            request.user.has_perm('{parent_app_name}.change_kontaktdaten')):
            raise PermissionDenied()
        else:
            kontaktdaten = Kontaktdaten.objects.filter(fachschaftuser=id)
            serializer = KontaktdatenSerializer(
                kontaktdaten,
                many=True,
                context={'request': request}
            )
            return Response(serializer.data)


# URLs

urlpatterns = format_suffix_patterns([
    path("fachschaftuser/<int:id>/", FachschaftUserDetail.as_view(), name="fachschaftuser-detail"),
    path("kontaktdaten/<int:id>/", KontaktdatenListByFachschaftUser.as_view(), name="kontaktdaten-detail"),
    path("user/<int:id>/", UserDetail.as_view(), name="user-detail"),
])



from django.contrib import admin

from .models import (
    FachschaftUser,
    Gender,
    Studienabschnitt,
    Studiengang,
    Kontaktdaten,
)

from .forms import (
    KontaktdatenForm,
)


@admin.register(FachschaftUser)
class FachschaftUserAdmin(admin.ModelAdmin):
    model = FachschaftUser
    fields = [
        "nickname",
        "user",
        "gender",
        "studiengang",
        "studienabschnitt",
    ]
    list_display = (
        "nickname",
        "__str__",
        "gender",
        "studiengang",
        "studienabschnitt",
    )
    search_fields = [
        "nickname",
        "user__first_name",
        "user__last_name",
    ]
    list_filter = (
        "gender",
        "studiengang",
        "studienabschnitt",
    )


@admin.register(Gender)
class GenderAdmin(admin.ModelAdmin):
    model = Gender
    fields = ["bezeichnung", "endung"]


@admin.register(Studiengang)
class StudiengangAdmin(admin.ModelAdmin):
    model = Studiengang
    fields = ["bezeichnung"]
    # If you don"t specify this, you will get a multiple select widget:
    filter_horizontal = ("studienabschnitt",)


admin.site.register(Studienabschnitt)


@admin.register(Kontaktdaten)
class KontaktdatenAdmin(admin.ModelAdmin):
    readonly_fields = [
        "fachschaftuser",
        "created",
        "adresse",
        "telefonnummer",
    ]
    fields = [
        "fachschaftuser",
        "created",
        "adresse",
        "telefonnummer",
        "gecheckt_von",
        "gecheckt_datum",
    ]
    list_display = [
        "fachschaftuser",
        "created",
        "adresse",
        "telefonnummer",
        "gecheckt_von",
        "gecheckt_datum",
    ]



from django.urls import path, include
from . import views, api

app_name = 'fsmedhro_core'

urlpatterns = [
    path('edit/', views.FachschaftUserEdit.as_view(), name='edit'),
    path('api/', include(api)),
    path(
        'kontaktdaten/add/',
        views.KontaktdatenAdd.as_view(),
        name='kontaktdaten_add',
    ),
    path(
        'kontaktdaten/verify/',
        views.KontaktdatenVerify.as_view(),
        name='kontaktdaten_verify',
    ),
    path('kontaktdaten/', views.KontaktdatenList.as_view(), name='kontaktdaten_list'),
    path('', views.FachschaftUserDetail.as_view(), name='detail'),
]



import logging

from datetime import datetime

from django.conf import settings
from django.contrib.auth import REDIRECT_FIELD_NAME
from django.contrib.auth.mixins import LoginRequiredMixin, UserPassesTestMixin
from django.core.mail import send_mail, EmailMessage
from django.forms.models import model_to_dict
from django.shortcuts import redirect, render, get_object_or_404
from django.views import View
from django.views.generic import ListView

from rest_framework.decorators import api_view
from rest_framework.response import Response
from rest_framework.reverse import reverse

from .forms import FachschaftUserForm, KontaktdatenForm
from .models import Studienabschnitt, Studiengang, Gender, FachschaftUser, Kontaktdaten


logger = logging.getLogger(__name__)


class FachschaftUserEdit(LoginRequiredMixin, View):
    template_name = "fsmedhro_core/user_edit.html"
    form_class = FachschaftUserForm

    def get(self, request):
        # FachschaftUser bereits vorhanden?
        if hasattr(request.user, 'fachschaftuser'):
            fu_form = self.form_class(instance=request.user.fachschaftuser)
        else:
            fu_form = self.form_class()

        context = {
            "fu_form": fu_form,
            "next": request.GET.get(REDIRECT_FIELD_NAME, ""),
        }

        return render(request, self.template_name, context)

    def post(self, request):
        if hasattr(request.user, 'fachschaftuser'):
            fu_form = self.form_class(
                data=request.POST,
                instance=request.user.fachschaftuser,
            )
        else:
            fu_form = self.form_class(
                data=request.POST,
            )

        if fu_form.is_valid():
            fachschaftuser = fu_form.save(commit=False)
            fachschaftuser.user = request.user
            fachschaftuser.save()

            # TODO: redirect zu next, falls vorhanden

            return redirect("fsmedhro_core:detail")
        else:
            context = {
                "fu_form": fu_form,
            }
            return render(request, self.template_name, context)


class FachschaftUserDetail(LoginRequiredMixin, View):
    template_name = "fsmedhro_core/user_detail.html"

    def get(self, request):
        if not hasattr(request.user, 'fachschaftuser'):
            return redirect("fsmedhro_core:edit")

        fachschaftuser = request.user.fachschaftuser

        context = {
            "fachschaftuser": fachschaftuser,
            "aktuelle_kontaktdaten": fachschaftuser.kontaktdaten.first(),
            "kontaktdaten": fachschaftuser.kontaktdaten.all(),
        }

        return render(request, self.template_name, context)


class KontaktdatenList(LoginRequiredMixin, UserPassesTestMixin, ListView):
    queryset = Kontaktdaten.objects.select_related('fachschaftuser__user').filter(gecheckt_von=None)
    context_object_name = 'kontaktdaten'
    template_name = 'fsmedhro_core/kontaktdaten_list.html'

    def test_func(self):
        return self.request.user.has_perm('fsmedhro_core.change_kontaktdaten')


class KontaktdatenVerify(LoginRequiredMixin, UserPassesTestMixin, View):
    def test_func(self):
        return self.request.user.has_perm('fsmedhro_core.change_kontaktdaten')

    def post(self, request):
        print(request.POST)
        k_id = request.POST.get('kontaktdaten_id', None)
        k = get_object_or_404(Kontaktdaten, pk=k_id)

        k.gecheckt_von = request.user
        k.gecheckt_datum = datetime.now()
        k.save()

        return redirect('fsmedhro_core:kontaktdaten_list')


class KontaktdatenAdd(LoginRequiredMixin, View):
    template_name = "fsmedhro_core/kontaktdaten_add.html"
    form_class = KontaktdatenForm

    def dispatch(self, request, *args, **kwargs):
        if not hasattr(request.user, 'fachschaftuser'):
            return redirect("fsmedhro_core:edit")
        else:
            return super().dispatch(request, *args, **kwargs)

    def get(self, request):
        context = {
            "kontaktdaten_form": self.form_class(),
        }

        return render(request, self.template_name, context)

    def post(self, request):
        kontaktdaten_form = self.form_class(data=request.POST)

        if kontaktdaten_form.is_valid():
            kontaktdaten = kontaktdaten_form.save(commit=False)
            kontaktdaten.save()

            return redirect("fsmedhro_core:detail")
        else:
            context = {
                "kontaktdaten_form": kontaktdaten_form,
            }

            return render(request, self.template_name, context)


class Rundmail(UserPassesTestMixin, View):
    template_name = "fsmedhro_core/rundmail.html"

    def test_func(self):
        return self.request.user.is_staff

    def get_context_data(self):
        context = {
            "studienabschnitte": Studienabschnitt.objects.all(),
            "studiengaenge": Studiengang.objects.all(),
            "gender": Gender.objects.all(),
        }
        return context

    def get(self, request):
        context = self.get_context_data()
        return render(request, self.template_name, context)

    def post(self, request):
        context = self.get_context_data()
        errors = []

        studiengaenge = list(map(int, request.POST.getlist("studiengang", [])))
        context["gew_studiengaenge"] = studiengaenge
        if not studiengaenge:
            errors.append("Es wurde kein Studiengang ausgewählt.")

        studienabschnitte = list(map(int, request.POST.getlist("studienabschnitt", [])))
        context["gew_studienabschnitte"] = studienabschnitte
        if not studienabschnitte:
            errors.append("Es wurde kein Studienabschnitt ausgewählt.")

        gender = list(map(int, request.POST.getlist("gender", [])))
        context["gew_gender"] = gender
        if not gender:
            errors.append("Es wurde kein Geschlecht ausgewählt.")

        betreff = request.POST.get("email_subject", "")
        context["betreff"] = betreff
        if not betreff:
            errors.append("Der Betreff fehlt.")

        text = request.POST.get("email_text", "")
        context["text"] = text
        if not text:
            errors.append("Der Text fehlt.")

        send_testmail = bool(request.POST.get("send_testmail", False))

        if errors:
            context["errors"] = errors
            return render(request, "fsmedhro_core/rundmail.html", context)
        elif send_testmail:
            anzahl_verschickt = send_mail(
                subject=betreff,
                message=text,
                from_email=settings.DEFAULT_FROM_EMAIL,
                recipient_list=[settings.DEFAULT_FROM_EMAIL],
                fail_silently=False,
            )
            logger.info(
                f"{request.user} verschickte Testmail\n" +
                f"{betreff=}\n" +
                f"{text=}"
            )
            context["testmail_verschickt"] = True
            context["anzahl_verschickt"] = anzahl_verschickt

            return render(request, self.template_name, context)
        else:
            empfaenger = FachschaftUser.objects.filter(
                studiengang__in=studiengaenge,
                studienabschnitt__in=studienabschnitte,
                gender__in=gender,
            ).order_by("user__email")
            empfaenger_adressen = [empf.user.email for empf in empfaenger]

            mails = EmailMessage(
                subject=betreff,
                body=text,
                bcc=empfaenger_adressen,
            )

            mails.send()
            logger.info(
                f"{request.user} verschickte Rundmail\n" +
                f"{betreff=}\n" +
                f"{text=}\n" +
                f"an" +
                ", ".join(empfaenger_adressen)
            )

            context["anzahl_verschickt"] = len(empfaenger)

            sicherheitsnachricht = (
                f"Die folgende Nachricht:\n\n" +
                f"Betreff: {betreff}\n" +
                f"Text:\n" +
                f"{text}\n\n" +
                f"wurde von {request.user} "
                f"an {len(empfaenger)} Personen verschickt:\n\n" +
                "\n".join(empfaenger_adressen)
            )
            send_mail(
                subject="Rundmail verschickt: " + betreff,
                message=sicherheitsnachricht,
                from_email=settings.DEFAULT_FROM_EMAIL,
                recipient_list=[settings.DEFAULT_FROM_EMAIL],
                fail_silently=False,
            )

            return render(request, self.template_name, context)



from django.contrib.auth.models import User
from django import forms

from .models import (
    FachschaftUser,
    Kontaktdaten,
)


class UserForm(forms.ModelForm):
    class Meta:
        model = User
        fields = ['first_name', 'last_name', 'email']


class FachschaftUserForm(forms.ModelForm):
    class Meta:
        model = FachschaftUser
        fields = [
            'nickname',
            'studiengang',
            'studienabschnitt',
            'gender',
        ]


class KontaktdatenForm(forms.ModelForm):
    class Meta:
        model = Kontaktdaten
        fields = [
            "fachschaftuser",
            "adresse",
            "telefonnummer",
        ]
        widgets = {
            "fachschaftuser": forms.HiddenInput(),
            "adresse": forms.Textarea(attrs={"rows": 3, "cols": 60}),
        }



import logging
from ldap3 import Server, Connection, ALL_ATTRIBUTES
from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib import messages

logger = logging.getLogger(__name__)
UserModel = get_user_model()


class LDAPUniRostock:
    HOST = 'ldaps://ldap.uni-rostock.de'
    BASE_DN = 'ou=people,o=uni-rostock,c=de'
    server = Server(HOST)

    def authenticate(self, request, username, password):
        username = username.lower()
        uid_qualifier = f'uid={username}'
        user_dn = f'{uid_qualifier},{self.BASE_DN}'

        logger.debug(f'Trying to authenticate with user_dn: {user_dn}')
        conn = Connection(self.server, user_dn, password)

        """
        Für die Authentifizierung versuchen wir uns hier an den LDAP-Server
        zu binden. Wenn es hierbei keine Exception gibt, dann ist der
        Benutzer gültig.
        """
        bind_successful = conn.bind()

        if bind_successful:
            """
            Wenn wir uns erfolgreich binden konnten, dann holen wir uns vom
            LDAP-Server alle Informationen um unseren eigenen Nutzer zu
            befüllen.
            """
            uid_search_term = f'({uid_qualifier})'
            conn.search(
                self.BASE_DN,
                uid_search_term,
                attributes=ALL_ATTRIBUTES
            )
            ldap_user = conn.entries[0]
            logger.debug(f'Got LDAP User: {ldap_user}')
        else:
            logger.debug('Could not authenticate. Unsuccessful bind.')
            messages.warning(
                request,
                'Authentifizierung bei der Uni fehlgeschlagen – '
                'bitte überprüfe dein Kürzel und Passwort.'
            )
            return None

        conn.unbind()

        if self.is_valid(ldap_user):
            try:
                user = UserModel.objects.get(username=username)
            except UserModel.DoesNotExist:
                """
                Wenn der Benutzer noch nicht existiert, dann erstellen wir
                einen:
                """
                user = UserModel(
                    username=ldap_user.uid.value,
                    password='DummyPassword',
                    is_staff=False,
                    is_superuser=False,
                )

                user.set_unusable_password()

                user.save()

            if user.is_active:
                user.email = ldap_user.mail.value
                user.first_name = ldap_user.givenName.value
                user.last_name = ldap_user.sn.value

                user.save()

            return user
        else:
            messages.warning(
                request,
                "Zutritt gibt's leider nur für Studenten "
                "der medizinischen Fakultät."
            )
            return None

    def is_valid(self, ldap_user):
        """
        Es werden nur Studenten von der Medizinischen Fakultät akzeptiert,
        außer die Accounts mit Ausnahmeregelungen in
        settings.FSRMED_AUTH_EXCEPTIONS.
        """
        try:
            logger.debug(
                f'Current auth exceptions: {settings.FSRMED_AUTH_EXCEPTIONS}'
            )
            if ldap_user.uid.value in settings.FSRMED_AUTH_EXCEPTIONS:
                return True
        except AttributeError:
            pass

        ldap_auth_filter = {
            'employeeType': {'s', 'si'},
            'uniRFaculty': {'03'},
            'gidNumber': {97},
        }

        """
        Es wird jedes Filterkriterium auf Gültigkeit überprüft.
        Der Wert vom ldap_user für die relevanten Schlüssel aus ldap_auth_filter muss
        in den dazugehörigen Werten aus ldap_auth_filter vorkommen.
        """
        for key, target_values in ldap_auth_filter.items():
            v = ldap_user[key].value
            invalid_user_warning_msg = (
                f'Invalid user {ldap_user.uid} because of {key} '
                f'(should be in {target_values} but is {v})'
            )

            if type(v) == int:
                vs = {v}
            elif type(v) == str:
                # Es könnte sein, dass ';' die verschiedenen employeeTypes trennt.
                vs = set(v.split(';'))
            else:
                vs = set(v)

            """
            Die Daten vom User müssen in allen Punkten mind. eine Übereinstimmung
            mit den Zieldaten haben, also z.B. wenn der Nutzer den employeeType
            'mi;si' hat (deswegen der split oben), dann soll der Nutzer sich einloggen
            können.
            """
            if not vs & target_values:
                logger.warning(invalid_user_warning_msg)
                return False
        else:
            return True

    def get_user(self, user_id):
        try:
            return UserModel.objects.get(pk=user_id)
        except UserModel.DoesNotExist:
            return None
